import numpy as np
import pandas as pd
from sklearn.datasets import load_breast_cancer
from sklearn.metrics import (
    average_precision_score,
    precision_recall_curve,
    precision_score,
)
from sklearn.model_selection import train_test_split

from scorepyo._utils import fast_numba_auc
from scorepyo.calibration import VanillaCalibrator
from scorepyo.models import EBMRiskScore
from scorepyo.ranking import LogOddsDensity


def test_end_2_end():
    # assert True

    data = load_breast_cancer()
    data_X, data_y = data.data, data.target

    X = pd.DataFrame(data=data_X, columns=data.feature_names)
    X["category"] = np.where(X["mean smoothness"] <= 0.1, "A", "B")
    y = pd.Series(data_y)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.25, random_state=0
    )

    X_test["category"] = "C"

    min_point_value = -2
    max_point_value = 3
    nb_max_features = 4

    ranker = LogOddsDensity()

    optim_method = fast_numba_auc

    scorepyo_model = EBMRiskScore(
        min_point_value=min_point_value,
        max_point_value=max_point_value,
        nb_max_features=nb_max_features,
        nb_additional_features=6,
        enumeration_maximization_metric=optim_method,
        ranker=ranker,
        calibrator=VanillaCalibrator(),
    )

    scorepyo_model.fit(
        X_train,
        y_train,
        X_calib=None,
        y_calib=None,
        categorical_features=["category"],
    )

    print(scorepyo_model.summary())

    y_proba = scorepyo_model.predict_proba(X_test)[:, 1].reshape(-1, 1)

    precision_recall_curve(y_test.astype(int), y_proba)
    average_precision = np.round(
        average_precision_score(y_test.astype(int), y_proba), 3
    )

    print(f"Average precision : \n{average_precision}")

    precision_test = precision_score(y_test.astype(int), y_proba > 0.5)

    print(f"Precision@0.5: \n{precision_test}")

    assert True



import pandas as pd
import pytest
from sklearn.datasets import load_breast_cancer


@pytest.fixture(scope="module")
def continuous_features():
    X = pd.DataFrame(columns=["A", "B"], data=[[0.2, 0.7], [0.1, 0.9], [7.8, -7]])
    return X


@pytest.fixture(scope="module")
def binary_features():
    X = pd.DataFrame(columns=["binary_A", "binary_B"], data=[[0, 1], [1, 1], [1, 1]])
    return X


@pytest.fixture(scope="module")
def df_info_binary_features():
    df_info = pd.DataFrame(
        columns=["feature", "binary_feature"],
        data=[["A", "binary_A"], ["B", "binary_B"]],
    )
    return df_info


@pytest.fixture(scope="module")
def mixed_features():
    X = pd.DataFrame(
        columns=["A", "B", "C"], data=[[0.2, 0.7, "1"], [0.1, 0.9, "2"], [7.8, -7, "2"]]
    )
    return X


@pytest.fixture(scope="module")
def continuous_target():
    y = pd.Series(data=[0.2, 0.7, 0.5])
    return y


@pytest.fixture(scope="module")
def binary_target():
    y = pd.Series(data=[0, 1, 1])
    return y


@pytest.fixture(scope="module")
def multiclass_target():
    y = pd.Series(data=[0, 1, 2])
    return y


@pytest.fixture(scope="module")
def breast_cancer_dataset():
    data = load_breast_cancer()
    return data


@pytest.fixture(scope="module")
def breast_cancer_features(breast_cancer_dataset):
    data = breast_cancer_dataset
    data_X = data.data
    X = pd.DataFrame(data=data_X, columns=data.feature_names)
    return X


@pytest.fixture(scope="module")
def breast_cancer_target(breast_cancer_dataset):
    data = breast_cancer_dataset
    return data.target



""" Class for calibrators
"""

from abc import ABC, abstractmethod
from typing import Any, List

import cvxpy as cp
import numpy as np
import pandas as pd
from numpy.random import default_rng

from scorepyo.exceptions import NonIncreasingProbabilities, NonProbabilityValues


class Calibrator(ABC):
    """Base Class for calibrator.

    RiskScore models will assign a score to each sample. Since the scores are a simple sum of few integers,
    we can enumerate all the possible values for scores. At each score value, there are negative and positive samples.
    Perfectly calibrated probabilities on a dataset would be proportion of positives at each score.
    However, there is a need for preserving the increasing trend between score and probability.
    For that, we optimize the logloss which naturally leads to calibrated probabilities, under ordering constraints
    The optimization is based on the counting of positive and negative samples at each possible score value.
    Args:
        ABC (_type_): _description_
    """

    def __init__(self, **kwargs):  # pylint: disable=W0613
        ...

    def calibrate(
        self, df_score: pd.DataFrame, min_sum_point: int, max_sum_point: int, **kwargs
    ) -> pd.DataFrame:
        """Funtion that takes a Dataframe of scores and binary target, and compute the associated probabilities for each sum of points

        It will use the _calibrate function defined in the child classes to compute these probabilities.

        Args:
            df_score (pd.DataFrame): DataFrame of scores and binary target
            min_sum_point (int): minimum possible sum of points
            max_sum_point (int): maximum possible sum of points

        Returns:
            pd.DataFrame: DataFrame containing the probability assigned to each possible score( or sum of points)
        """

        # computing the proportion of positive samples for each possible score value
        # This corresponds to the natural probability that would be the most calibrated without the reordering constraints

        df_score_proba_association = df_score.groupby("score")["target"].mean()
        df_score_proba_association.columns = ["proba"]

        # computing the full range of possible scores from min to max
        full_index = np.arange(min_sum_point, max_sum_point + 1e-3)

        # Reindexing the dataframe with the missing scores from the data
        df_score_proba_association = df_score_proba_association.reindex(full_index)

        # Interpolating the natural probability linearly
        df_score_proba_association = df_score_proba_association.interpolate(
            method="linear"
        )

        df_reordering = pd.DataFrame(df_score_proba_association.copy())

        # Count cardinality of each score
        df_reordering["count"] = df_score.groupby("score")["target"].count()
        df_reordering["count"] = df_reordering["count"].fillna(0)
        df_reordering["target"] = df_reordering["target"].fillna(0)

        # Adjust probability to have an optimized logloss and calibration
        df_cvx = df_reordering.copy()

        df_cvx["positive_count"] = df_cvx["target"] * df_cvx["count"]
        df_cvx["negative_count"] = df_cvx["count"] - df_cvx["positive_count"]

        df_cvx["positive_proba"] = df_cvx["positive_count"] / df_cvx["count"].sum()
        df_cvx["negative_proba"] = df_cvx["negative_count"] / df_cvx["count"].sum()

        # Call _calibrate function of child class
        list_calibrated_proba = self._calibrate(df_cvx, **kwargs)

        _check_is_probability = all(
            [(0 <= p) and (p <= 1) for p in list_calibrated_proba]
        )

        if not _check_is_probability:
            raise NonProbabilityValues(
                f"Probability values outputed by calibrator are not between 0 and 1. \n Probability list : {list_calibrated_proba}."
            )
        _check_increasing = all(
            [
                list_calibrated_proba[i] <= list_calibrated_proba[i + 1]
                for i in range(len(list_calibrated_proba) - 1)
            ]
        )
        if not _check_increasing:
            raise NonIncreasingProbabilities(
                f"Probability values outputed by calibrator must be increasing. \n Probability list : {list_calibrated_proba}."
            )

        # Get the optimized value for the probabilities
        df_reordering["sorted_proba"] = list_calibrated_proba
        return df_reordering

    @abstractmethod
    def _calibrate(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> List[float]:
        raise NotImplementedError


class VanillaCalibrator(Calibrator):
    """Vanilla calibrator that simply optimizes the logloss under ordering constraints.

    Given a list of increasing scores, and an associated count of negative and positive samples,
    this calibrator computes the probabilities by optimizing the logloss on the whole dataset,
    and respecting the ordering of probabilities according to scores.

    This calibrator should be favored when calibrating on a large dataset.

    """

    def __init__(self, **kwargs):
        ...

    def _calibrate(self, df_cvx: pd.DataFrame, **_kwargs) -> List[float]:

        # Declare the list of probabilities to be set as variables
        list_proba = [cp.Variable(1) for _ in df_cvx.index]

        # Compute total size of samples to normalize the logloss
        total_count = df_cvx["count"].sum()

        constraints = []
        for p in list_proba:
            constraints.append(p >= 0)
            constraints.append(p <= 1)
        for i in range(1, len(list_proba)):
            # TODO : Put the threshold away and combine all similar scores into 1
            constraints.append(list_proba[i] - list_proba[i - 1] - 1e-3 >= 0)

        # # Compute number of positive and negative samples at each score value
        positive_sample_count = (
            df_cvx["target"].values * df_cvx["count"].values
        ).astype(int)
        negative_sample_count = df_cvx["count"].values - positive_sample_count
        list_expression = [
            -cp.log(p) * w_pos - cp.log(1 - p) * w_neg
            for p, w_pos, w_neg in zip(
                list_proba, positive_sample_count, negative_sample_count
            )
        ]
        objective = cp.Minimize(cp.sum(list_expression) / total_count)  # Objectiv

        problem = cp.Problem(objective, constraints)

        _ = problem.solve(verbose=False)

        return [p.value[0] for p in list_proba]


class BootstrappedCalibrator(Calibrator):
    """Bootstrapped calibrator that optimizes the logloss under ordering constraints on different bootstrapped sets.

    Given an original dataset, this calibrator bootstraps several times other dataset and finds probability that optimize the logloss on all datasets,
    still respecting the probability ordering by score.
    This BootstrappedCalibrator class has two modes, it can either:
    - optimize the average logloss across all bootstrapped datasets
    - optimize the worse logloss among dataset

    The latter will lead to worse logloss on the training dataset, but more robust logloss on the test set if it's similar to the distribution on the training dataset.
    The BootstrappedCalibrator should be favored when calibrating on a small dataset.

    """

    def __init__(self, nb_experiments: int = 20, method: str = "average", **_kwargs):
        """
        Args:
            nb_experiments (int, optional): Number of bootstrapped datasets used to optimize the logloss. Defaults to 20.
            method (str, optional): Indicator of average mode or worst_case mode. Defaults to "average".
        """
        self.nb_experiments = nb_experiments
        if method not in ("average", "worst_case"):
            raise NotImplementedError
        self.method = method

    def _calibrate(self, df_cvx: pd.DataFrame, **_kwargs) -> List[float]:
        list_proba_multinomial = list(df_cvx["positive_proba"].values)
        list_proba_multinomial += list(df_cvx["negative_proba"].values)

        rng = default_rng()
        nb_experiment = self.nb_experiments
        size_sample = df_cvx["count"].sum()

        # Bootstrapping here corresponds to samples drawn from a multinomial distribution
        bootstrap_samples = rng.multinomial(
            n=size_sample, pvals=list_proba_multinomial, size=nb_experiment
        )

        positive_count_samples = bootstrap_samples[:, : len(df_cvx)].T
        negative_count_samples = bootstrap_samples[:, len(df_cvx) :].T
        positive_col_bootstrap = [f"positive_count_{i}" for i in range(nb_experiment)]
        negative_col_bootstrap = [f"negative_count_{i}" for i in range(nb_experiment)]
        df_positive = pd.DataFrame(
            positive_count_samples, columns=positive_col_bootstrap, index=df_cvx.index
        )
        df_negative = pd.DataFrame(
            negative_count_samples, columns=negative_col_bootstrap, index=df_cvx.index
        )

        df_cvx = pd.concat([df_cvx, df_positive, df_negative], axis=1)

        # Declare the list of probabilities to be set as variables
        list_proba = [cp.Variable(1) for _ in df_cvx.index]

        # Compute total size of samples to normalize the logloss
        total_count = df_cvx["count"].sum()

        # Compute the logloss at each score in a list in order to sum it later
        # the logloss at each score is simple as all samples will have the same
        # probability p. for all positive samples, add -log(p), for all negative samples add -log(1-p)
        list_bootstrap_sample_objective = []
        for i in range(nb_experiment):
            positive_sample_count = df_cvx[f"positive_count_{i}"].values
            negative_sample_count = df_cvx[f"negative_count_{i}"].values
            list_expression = [
                -cp.log(p) * w_pos - cp.log(1 - p) * w_neg
                for p, w_pos, w_neg in zip(
                    list_proba, positive_sample_count, negative_sample_count
                )
            ]
            list_bootstrap_sample_objective.append(list_expression.copy())
        # objective = cp.Minimize(cp.sum(list_expression) / total_count)  # Objectiv

        constraints = []
        for p in list_proba:
            constraints.append(p >= 0)
            constraints.append(p <= 1)
        for i in range(1, len(list_proba)):
            # TODO : Put the threshold away and combine all similar scores into 1
            constraints.append(list_proba[i] - list_proba[i - 1] - 1e-3 >= 0)

        # If worst_case mode is chosen, optimize an auxiliary variable
        # that will correspond to the worst log loss among one of the bootstrapped datasets
        if self.method == "worst_case":
            worse_log_loss = cp.Variable(1)
            for single_bootstrap_logloss in list_bootstrap_sample_objective:
                constraints.append(
                    worse_log_loss >= cp.sum(single_bootstrap_logloss) / total_count
                )
            objective = cp.Minimize(worse_log_loss)

        # else if average mode, optimize the mean logloss across all bootstrapped datasets
        elif self.method == "average":
            objective = cp.Minimize(
                cp.sum(
                    [
                        cp.sum(single_bootstrap_logloss) / total_count
                        for single_bootstrap_logloss in list_bootstrap_sample_objective
                    ]
                )
            )

            objective = cp.Minimize(cp.sum(list_expression) / total_count)  # Objectiv

        else:
            raise Exception

        problem = cp.Problem(objective, constraints)

        _ = problem.solve(verbose=False)

        return [p.value[0] for p in list_proba]



""" Definition of the different classes for the Error management """
from __future__ import annotations

import pandera as pa
import pandera.extensions as extensions
from pandas.api.types import is_numeric_dtype


@extensions.register_check_method()  # decorator needed by pandera
def is_numeric(pandas_obj):
    """function that returns true if series dtype is numeric

    The decorator is meant to be

    Args:
        pandas_obj (Pandas.Series): series with the datatype check

    Returns:
        bool: True if data type is numeric, False if not
    """
    return is_numeric_dtype(pandas_obj)


def NumericCheck():
    """Checks that a pandas Series has a numeric data type

    THis checks is to ensure that the rolling aggregation of pandas will run smoothly afterward

    Returns:
        pandera.Check: Pandera check
    """
    return pa.Check.is_numeric(
        title="Numeric type check",
        description="Check that the column is a numeric type to ensure pandas rolling aggregation will work",
    )


# General errors


class MissingColumnError(Exception):
    """Error raised when column is missing in a pd.DataFrame"""

    pass


class NegativeValueError(Exception):
    """Error raised when quantity value is not a strictly positive float"""

    pass


class NonIntegerValueError(Exception):
    """Error raised when value is not an integer"""

    pass


class MinPointOverMaxPointError(Exception):
    """Error raised when min point value of ScoreCard model
    is over max point value"""

    pass


class NonBooleanValueError(Exception):
    """Error raised when value is not boolean"""

    pass


class NonProbabilityValues(Exception):
    """Error raised when a probability value returned by a calibrator is not valid (between 0 and 1)"""

    pass


class NonIncreasingProbabilities(Exception):
    """Error raised when list of probabilities returned by a calibrator is not increasing"""

    pass



"""Class for binarizers:
- EBM-based automatic binarizer
- TODO : Quantile binarizer
"""
import numbers
import time
import warnings
from math import floor, log10
from typing import List, Optional, Protocol, Union

import numpy as np
import pandas as pd
import pandera as pa
from interpret.glassbox import ExplainableBoostingClassifier
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import OneHotEncoder

from scorepyo.exceptions import (
    MissingColumnError,
    NegativeValueError,
    NonBooleanValueError,
    NonIntegerValueError,
    NumericCheck,
)


class BinarizerProtocol(Protocol):
    """Protocol to respect for future and custom binarizer"""

    def fit(self, X, y, *args, **kwargs):
        ...

    def transform(self, X, **kwargs) -> pd.DataFrame:
        ...

    def get_info(self) -> pd.DataFrame:
        ...


class EBMBinarizer:
    """
    Class for automatic feature binarizer based on EBM.

    This class uses Explainable Boosting Machine (EBM) that are part of the General Additive Model (GAM) family.
    EBM will compute for each feature a tree. The final prediction will be made by summing up the tree value for each feature.
    Since it's single-feature tree, it actually is a piecewise constant scalar function.
    For each interval where the single-feature tree yields a constant contribution for the prediction, this class computes a binary feature.

    Attributes
    ----------
    max_number_binaries_by_features : int
        maximum number of binary features to compute by feature
    keep_negative : bool
        indicator to keep features that decrease predicted probability
    _ebm: ExplainableBoostingClassifier
        EBM classifier that is fitted and used to binarize feature
    _one_hot_encoder: OneHotEncoder
        Scikit-learn one-hot encoder for categorical features
    _categorical_features: list(str)
        list of categorical features
    _continuous_features: list(str)
        list of continuous features
    _to_exclude_features: list(str)
        list of features to exclude from binarizing


    Methods
    -------
    fit(X, y, categorical_features="auto", to_exclude_features=None))
        fits the EBM model on X,y and the one-hot encoder on the categorical features

    transform(X)
        transforms features in X into binarized features based on previously fitted EBM and one-hot encoder
    """

    def __init__(
        self,
        max_number_binaries_by_features: int = 2,
        keep_negative: bool = True,
    ):
        """
        Args:
            max_number_binaries_by_features (int, optional): maximum number of binary features to compute by feature. Defaults to 2.
            keep_negative (bool, optional): indicator to keep features that decrease predicted probability. Defaults to False.
        """

        if max_number_binaries_by_features <= 0:
            raise NegativeValueError(
                f"max_number_binaries_by_features must be a strictly positive integer. \n {max_number_binaries_by_features} is not positive."
            )
        if not isinstance(
            max_number_binaries_by_features,
            numbers.Integral,
        ):
            raise NonIntegerValueError(
                f"max_number_binaries_by_features must be a strictly positive integer. \n {max_number_binaries_by_features} is not an integer."
            )

        self.max_number_binaries_by_features = max_number_binaries_by_features

        if not isinstance(keep_negative, bool):
            raise NonBooleanValueError(
                f"keep_negative attribute must be a boolean. \n {keep_negative} is not a boolean"
            )

        self.keep_negative: bool = keep_negative

        # Creation of underlying EBM to extract binary feature
        # interactions to 0 to prevent pairwise interaction feature
        # TODO : Include pairwise interaction into featurizer?
        # max_bins controls the number of split for each single-feature tree
        self._ebm: BaseEstimator = ExplainableBoostingClassifier(
            interactions=0,
            max_bins=self.max_number_binaries_by_features + 2,
            min_samples_leaf=10,
        )

        # One-hot encoder that imputes infrequent_if_exist for unknown categories
        # it allows only the 10 most frequent categories, in order to not create too many columns
        # for high-cardinality categories
        self._one_hot_encoder: BaseEstimator = OneHotEncoder(
            handle_unknown="infrequent_if_exist",
            # max_categories=10,
            sparse=False,
        )

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        categorical_features: Union[list[str], str] = "auto",
        to_exclude_features: Optional[list[str]] = None,
    ):
        """Fit function of binarizer

        This functions fits the EBM on X,y and the one-hot encoder on X.
        It stores the categorical, continuous and to exclude columns.

        Args:
            X (pandas.Dataframe): Dataframe of features to use to fit the binarizer
            y (pandas.Series): target values
            categorical_features (list(str), optional): list of categorical features to one-hot encode.
            Defaults to "auto" to automatically detect categorical features.
            to_exclude_features (list(str), optional): List of features to leave as is.
            Defaults to None.
        """
        start_time = time.time()

        # TODO Use pandera
        if categorical_features == "auto":
            self._categorical_features = X.select_dtypes(
                include=[
                    "category",
                    "object",
                    "bool",
                ]
            ).columns
        else:
            not_present_categorical_features = set(categorical_features) - set(
                X.columns
            )
            if len(not_present_categorical_features) > 0:
                raise MissingColumnError(
                    f"{not_present_categorical_features} are not in columns of X."
                )

            self._categorical_features = categorical_features
        if to_exclude_features is None:
            self._to_exclude_features = []
        else:
            not_present_to_exclude_features = set(to_exclude_features) - set(X.columns)
            if len(not_present_to_exclude_features) > 0:
                raise MissingColumnError(
                    f"{not_present_to_exclude_features} are not in columns of X."
                )
            self._to_exclude_features = to_exclude_features

        self._categorical_features = [
            c for c in self._categorical_features if c not in self._to_exclude_features
        ]
        self._continuous_features = [
            c
            for c in X.columns
            if (c not in self._categorical_features)
            and (c not in self._to_exclude_features)
        ]
        cond_list = [
            [c in self._continuous_features for c in X.columns],
            [c in self._categorical_features for c in X.columns],
        ]
        choice_list = ["continuous", "nominal"]
        ebm_feature_type_param = np.select(
            condlist=cond_list, choicelist=choice_list, default=None  # type: ignore[arg-type]
        )
        self._ebm.set_params(feature_types=ebm_feature_type_param)
        self._ebm.fit(X, y)

        if len(self._categorical_features) > 0:
            self._one_hot_encoder.fit(X[self._categorical_features])

        ebm_global = self._ebm.explain_global(name="EBM")

        list_scores: List[Optional[float]] = []
        list_lower_threshold: List[Optional[float]] = []
        list_upper_threshold: List[Optional[float]] = []
        list_category_value: List[Optional[str]] = []
        list_original_column: List[Optional[str]] = []
        list_binary_feature_names: List[Optional[str]] = []
        list_columns = []
        list_feature_type: List[Optional[str]] = []
        list_density: List[Optional[float]] = []

        if len(self._categorical_features) > 0:
            dict_categorical_features = dict()

        # For each continuous feature, we look at their single-feature tree
        for (
            i,
            feature_name,
        ) in enumerate(ebm_global.data()["names"]):
            if feature_name in self._continuous_features:

                # Get EBM info for current feature
                dict_EBM_info_feature = ebm_global.data(i)

                # Get number of plateaus for the corresponding tree
                number_plateau = len(ebm_global.data(i)["scores"])

                # For each plateau, extract lower and upper bound of each interval
                for j in range(number_plateau):
                    contrib = dict_EBM_info_feature["scores"][j]

                    # If logodd contribution of the feature is negative,
                    # check if we should keep negative contribution to probability
                    if (contrib < 0) and (not self.keep_negative):
                        continue

                    # Store lower and upper bound of interval
                    threshold_lower = dict_EBM_info_feature["names"][j]
                    threshold_upper = dict_EBM_info_feature["names"][j + 1]

                    # Round precision for column name
                    # needed when difference upper and lower bound is below 2 digits

                    # try:
                    difference = threshold_upper - threshold_lower
                    # except:
                    #     # Debug
                    #     print(feature_name)
                    #     print(dict_EBM_info_feature)
                    #     assert False

                    needed_round_precision = -floor(log10(abs(difference)))
                    needed_round_precision = max(needed_round_precision, 2)

                    # Case of the first plateau, only have an upper bound for current feature value
                    if j == 0:
                        col_name = f"{feature_name} < {np.round(threshold_upper,needed_round_precision)}"

                        list_columns.append(
                            pd.DataFrame(
                                data=(X[feature_name] < threshold_upper).values,
                                columns=[col_name],
                            )
                        )
                        threshold_lower = np.nan

                    # Case of the last plateau, only have a lower bound for current feature value
                    elif j == number_plateau - 1:
                        col_name = f"{feature_name} >= {np.round(threshold_lower,needed_round_precision)}"
                        list_columns.append(
                            pd.DataFrame(
                                data=(X[feature_name] >= threshold_lower).values,
                                columns=[col_name],
                            )
                        )
                        threshold_upper = np.nan

                    # Case for other inbetween plateaus
                    else:
                        col_name = f"{np.round(threshold_lower,needed_round_precision)} <= {feature_name} < {np.round(threshold_upper,needed_round_precision)}"
                        list_columns.append(
                            pd.DataFrame(
                                data=(
                                    (X[feature_name] < threshold_upper)
                                    & (X[feature_name] >= threshold_lower)
                                ).values,
                                columns=[col_name],
                            )
                        )

                    list_density.append(list_columns[-1][col_name].sum())
                    list_feature_type.append("continuous")

                    list_lower_threshold.append(threshold_lower)
                    list_upper_threshold.append(threshold_upper)

                    list_category_value.append(None)

                    list_scores.append(contrib)
                    list_original_column.append(feature_name)
                    list_binary_feature_names.append(col_name)
            else:
                if (len(self._categorical_features) > 0) and (
                    feature_name in self._categorical_features
                ):
                    # Get EBM info for current feature
                    dict_EBM_info_feature = ebm_global.data(i)

                    # # Get number of plateaus for the corresponding tree
                    # number_categories = len(ebm_global.data(i)["names"])

                    feature_name = ebm_global.data()["names"][i]

                    # list_binary_features_name = []
                    for category_value, score, density in zip(
                        ebm_global.data(i)["names"],
                        ebm_global.data(i)["scores"],
                        ebm_global.data(i)["density"]["scores"],
                    ):
                        # If category value is of type int or float
                        if isinstance(category_value, int) or isinstance(
                            category_value, float
                        ):
                            # Test if int or float
                            if int(category_value) == float(category_value):
                                # it's a int
                                category_value = int(category_value)
                            else:
                                # it's a float
                                category_value = float(category_value)

                        binary_feature_name = "_".join(
                            [str(feature_name), str(category_value)]
                            # [feature_name, str(category_value)]
                        )
                        dict_categorical_features[binary_feature_name] = {
                            "score": score,
                            "density": density,
                        }

        # TODO : Put logodds of ebm for categorical features
        if len(self._categorical_features) > 0:
            # One-hot encode categorical features
            one_hot_categorical_columns = self._one_hot_encoder.get_feature_names_out()
            list_columns.append(
                pd.DataFrame(
                    data=(
                        self._one_hot_encoder.transform(X[self._categorical_features])
                    ),
                    columns=one_hot_categorical_columns,
                )
            )

            list_binary_feature_names.extend(one_hot_categorical_columns)

            # Add info for info dataframe
            for name_out in self._one_hot_encoder.get_feature_names_out():
                list_scores.append(dict_categorical_features[name_out]["score"])
                list_density.append(dict_categorical_features[name_out]["density"])
                list_feature_type.append("categorical")
                list_lower_threshold.append(None)
                list_upper_threshold.append(None)
                list_category_value.append(name_out.split("_")[-1])
                list_original_column.append("_".join(name_out.split("_")[:-1]))

        # Copy features to exclude from binarizing
        if len(self._to_exclude_features) > 0:
            list_columns.append(X[self._to_exclude_features].copy())

            for name_out in self._to_exclude_features:
                list_scores.append(0)
                list_density.append(0)
                list_feature_type.append(None)  # type: ignore
                list_lower_threshold.append(None)
                list_upper_threshold.append(None)
                list_category_value.append(None)
                list_original_column.append(name_out)

        # concat all created columns at the end (vs on the fly) for performance
        X_binarized = pd.concat(list_columns, axis=1)

        # convert to 1/0
        with warnings.catch_warnings():
            # Setting values in-place is fine, ignore the warning in Pandas >= 1.5.0
            # This can be removed, if Pandas 1.5.0 does not need to be supported any longer.
            # See also: https://stackoverflow.com/q/74057367/859591
            warnings.filterwarnings(
                "ignore",
                category=FutureWarning,
                message=(
                    ".*will attempt to set the values inplace instead of always setting a new array. "
                    "To retain the old behavior, use either.*"
                ),
            )
            X_binarized.loc[:, list_binary_feature_names] = X_binarized.loc[
                :, list_binary_feature_names
            ].astype(int)

        # Regroup information about binary features created in the info dataframe
        # this dataframe contains:
        # - the list for each binary feature of log odds contribution
        # - the list for each binary feature of lower threshold used
        # - the list for each binary feature of upper threshold used
        # - the list for each binary feature of the original feature it came from

        df_score_feature = pd.DataFrame(
            index=X_binarized.columns,
            columns=[
                "log_odds",
                "lower_threshold",
                "upper_threshold",
                "category_value",
                "feature",
                "type",
                "density",
            ],
            data=np.array(
                [
                    list_scores,
                    list_lower_threshold,
                    list_upper_threshold,
                    list_category_value,
                    list_original_column,
                    list_feature_type,
                    list_density,
                ]
            ).T,
        )

        # the intercept is the basis log odd value of the EBM without adding/substracting
        # the different log odds contribution of each feature
        row_intercept = pd.DataFrame(
            index=["intercept"],
            columns=df_score_feature.columns,
            data=[[self._ebm.intercept_[0], None, None, None, "intercept", None, None]],
        )
        df_score_feature = pd.concat(
            [
                df_score_feature,
                row_intercept,
            ],
            axis=0,
        )
        df_score_feature.index.name = "binary_feature"

        self.df_score_feature = df_score_feature.copy()
        self._fit_time = time.time() - start_time
        # return (
        #     X_binarized,
        #     df_score_feature,
        # )

    # TODO : Provide function that give the quality of the binarizer via the learning metrics of ebm

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if not getattr(self._ebm, "has_fitted_", False):
            raise NotFittedError("AutomaticFeatureBinarizer has not been fitted.")

        dict_check_continuous_features = {
            c: pa.Column(checks=[NumericCheck()]) for c in self._continuous_features
        }

        dict_check_other_features = {
            c: pa.Column()
            for c in self._categorical_features + self._to_exclude_features
        }

        dataframe_schema = pa.DataFrameSchema(
            {
                **dict_check_continuous_features,
                **dict_check_other_features,
            },
            strict=True,  # Enable check of other columns in the dataframe
        )

        dataframe_schema.validate(X)

        list_columns = []
        list_binary_feature_names = []
        for (
            binary_feature_name,
            row,
        ) in self.df_score_feature.iterrows():
            feature_name = row["feature"]
            lower_threshold = row["lower_threshold"]
            upper_threshold = row["upper_threshold"]
            # category_value = row["category_value"]
            type = row["type"]

            if type == "continuous":
                list_binary_feature_names.append(binary_feature_name)
                if lower_threshold is np.nan:
                    list_columns.append(
                        pd.DataFrame(
                            data=(X[feature_name] < upper_threshold).values,
                            columns=[binary_feature_name],
                        )
                    )
                elif upper_threshold is np.nan:
                    list_columns.append(
                        pd.DataFrame(
                            data=(X[feature_name] >= lower_threshold).values,
                            columns=[binary_feature_name],
                        )
                    )
                else:
                    list_columns.append(
                        pd.DataFrame(
                            data=(
                                (X[feature_name] < upper_threshold)
                                & (X[feature_name] >= lower_threshold)
                            ).values,
                            columns=[binary_feature_name],
                        )
                    )

        if len(self._categorical_features) > 0:
            # One-hot encode categorical features
            one_hot_categorical_columns = self._one_hot_encoder.get_feature_names_out()
            list_columns.append(
                pd.DataFrame(
                    data=(
                        self._one_hot_encoder.transform(X[self._categorical_features])
                    ),
                    columns=one_hot_categorical_columns,
                )
            )

            list_binary_feature_names.extend(one_hot_categorical_columns)

        # Copy features to exclude from binarizing
        if len(self._to_exclude_features) > 0:
            list_columns.append(X[self._to_exclude_features].copy())

        # concat all created columns at the end (vs on the fly) for performance
        X_binarized = pd.concat(list_columns, axis=1)

        # convert to 1/0
        with warnings.catch_warnings():
            # Setting values in-place is fine, ignore the warning in Pandas >= 1.5.0
            # This can be removed, if Pandas 1.5.0 does not need to be supported any longer.
            # See also: https://stackoverflow.com/q/74057367/859591
            warnings.filterwarnings(
                "ignore",
                category=FutureWarning,
                message=(
                    ".*will attempt to set the values inplace instead of always setting a new array. "
                    "To retain the old behavior, use either.*"
                ),
            )
            X_binarized.loc[:, list_binary_feature_names] = X_binarized.loc[
                :, list_binary_feature_names
            ].astype(int)

            return X_binarized

    def get_info(self) -> pd.DataFrame:
        return self.df_score_feature



"""Fast Numba ROC AUC

credit : https://github.com/diditforlulz273/fastauc/blob/main/fastauc/fast_auc.py
"""

import numba
import numpy as np
import numpy.typing as npt


def fast_numba_auc(
    y_true: npt.NDArray, y_score: npt.NDArray, sample_weight: npt.NDArray = None
) -> float:
    """a function to calculate AUC via python + numba.
    Args:
        y_true (np.array): 1D numpy array as true labels.
        y_score (np.array): 1D numpy array as probability predictions.
        sample_weight (np.array): 1D numpy array as sample weights, optional.
    Returns:
        AUC score as float
    """
    if sample_weight is None:
        return fast_numba_auc_nonw(y_true=y_true, y_score=y_score)
    else:
        return fast_numba_auc_w(
            y_true=y_true, y_score=y_score, sample_weight=sample_weight
        )


@numba.njit
def trapezoid_area(x1: float, x2: float, y1: float, y2: float) -> float:
    dx = x2 - x1
    dy = y2 - y1
    return dx * y1 + dy * dx / 2.0


@numba.njit
def fast_numba_auc_nonw(y_true: npt.NDArray, y_score: npt.NDArray) -> float:
    y_true = y_true == 1

    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]

    prev_fps = 0
    prev_tps = 0
    last_counted_fps = 0
    last_counted_tps = 0
    auc = 0.0
    for i in range(len(y_true)):
        tps = prev_tps + y_true[i]
        fps = prev_fps + (1 - y_true[i])
        if i == len(y_true) - 1 or y_score[i + 1] != y_score[i]:
            auc += trapezoid_area(last_counted_fps, fps, last_counted_tps, tps)
            last_counted_fps = fps
            last_counted_tps = tps
        prev_tps = tps
        prev_fps = fps
    return auc / (prev_tps * prev_fps)


@numba.njit
def fast_numba_auc_w(
    y_true: npt.NDArray, y_score: npt.NDArray, sample_weight: npt.NDArray
) -> float:
    y_true = y_true == 1

    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    sample_weight = sample_weight[desc_score_indices]

    prev_fps = 0
    prev_tps = 0
    last_counted_fps = 0
    last_counted_tps = 0
    auc = 0.0
    for i in range(len(y_true)):
        weight = sample_weight[i]
        tps = prev_tps + y_true[i] * weight
        fps = prev_fps + (1 - y_true[i]) * weight
        if i == len(y_true) - 1 or y_score[i + 1] != y_score[i]:
            auc += trapezoid_area(last_counted_fps, fps, last_counted_tps, tps)
            last_counted_fps = fps
            last_counted_tps = tps
        prev_tps = tps
        prev_fps = fps
    return auc / (prev_tps * prev_fps)



# mypy: ignore-errors
"""Classes to create and fit risk-score type model.
"""
import itertools
import numbers
import time
from abc import abstractmethod
from typing import Any, Optional

import dask.array as da
import numpy as np
import pandas as pd
import pandera as pa
from sklearn.exceptions import NotFittedError

from scorepyo._utils import fast_numba_auc
from scorepyo.binarizers import BinarizerProtocol, EBMBinarizer
from scorepyo.calibration import Calibrator, VanillaCalibrator
from scorepyo.exceptions import (
    MinPointOverMaxPointError,
    NegativeValueError,
    NonIntegerValueError,
)
from scorepyo.ranking import OMPRank, Ranker


class _BaseRiskScore:
    """
    Base class for risk score type model.

    This class provides common functions for risk score type model, no matter the way points and binary features are designed.
    It needs common attributes, such as minimum/maximum point value for each binary feature and number of selected binary feature.


    Attributes
    ----------
    nb_max_features : int
        maximum number of binary features to compute by feature
    min_point_value : int
        minimum points to assign for a binary feature
    max_point_value: ExplainableBoostingClassifier
        maximum points to assign for a binary feature



    Methods
    -------
    @staticmethod
    _predict_proba_score(score, intercept):
        computes the value of the logistic function at score+intercept value

    @abstractmethod
    fit(self, X, y):
        function to be implemented by child classes. This function should create the feature-point card and score card attributes.

    predict(self, X, proba_threshold=0.5):
        function that predicts the positive/negative class by using a threshold on the probability given by the model

    predict_proba(self, X):
        function that predicts the probability of the positive class

    summary(self):
        function that prints the feature-point card and score card of the model
    """

    _DESCRIPTION_COL = "Description"
    """Column name for binary feature in the risk score summary"""

    _POINT_COL = "Point(s)"
    """Column name for points in the risk score summary"""

    _FEATURE_COL = "Feature"
    """Column name for original feature in the risk score summary"""

    def __init__(
        self,
        nb_max_features: int = 4,
        min_point_value: int = -2,
        max_point_value: int = 3,
        # df_info: Optional[pd.DataFrame] = None,
    ):
        """
        Args:
            nb_max_features (int): Number of maximum binary features to be selected
            min_point_value (int, optional): Minimum point assigned to a binary feature. Defaults to -2.
            max_point_value (int, optional): Maximum point assigned to a binary feature. Defaults to 3.
        """
        if nb_max_features <= 0:
            raise NegativeValueError(
                f"nb_max_features must be a strictly positive integer. \n {nb_max_features} is not positive."
            )
        if not isinstance(nb_max_features, numbers.Integral):
            raise NonIntegerValueError(
                f"nb_max_features must be a strictly positive integer. \n {nb_max_features} is not an integer."
            )
        self.nb_max_features: int = nb_max_features

        if not isinstance(min_point_value, numbers.Integral):
            raise NonIntegerValueError(
                f"min_point_value must be an integer. \n {min_point_value} is not an integer."
            )

        if not isinstance(max_point_value, numbers.Integral):
            raise NonIntegerValueError(
                f"max_point_value must be an integer. \n {max_point_value} is not an integer."
            )
        if min_point_value > max_point_value:
            raise MinPointOverMaxPointError(
                f"max_point_value must be greater than or equal to min_point_value. \n {max_point_value=} < {min_point_value=} ."
            )
        self.min_point_value: int = min_point_value
        self.max_point_value: int = max_point_value

        # self._df_info: Optional[pd.DataFrame]

        # if df_info is not None:
        #     dataframe_schema = pa.DataFrameSchema(
        #         {
        #             "binary_feature": pa.Column(),
        #             "feature": pa.Column(),
        #         },
        #         strict=False,  # Disable check of other columns in the dataframe
        #     )

        #     dataframe_schema.validate(df_info)
        #     self._df_info = df_info.copy()
        # else:
        #     self._df_info = None

        self.feature_point_card: Optional[pd.DataFrame] = None
        self.score_card: Optional[pd.DataFrame] = None

    @abstractmethod
    def fit(
        self, X: pd.DataFrame, y: pd.Series, *args: Any, **kwargs: Any
    ) -> Optional[NotImplementedError]:
        """Functions that creates the feature-point card and score card

        Must be defined for each child class

        Args:
            X (pandas.DataFrame): binary feature dataset
            y (pandas.Series): binary target

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def predict(self, X: pd.DataFrame, proba_threshold: float = 0.5) -> np.ndarray:
        """Predicts binary class based on probabillity threshold

        Afer computing the probability for each sample,

        Args:
            X (pd.DataFrame): _description_
            proba_threshold (float, optional): probability threshold for binary classification. Defaults to 0.5.

        Returns:
            nbarray of shape (n_samples,): predicted class based on predicted probability and threshold
        """
        if (self.feature_point_card is None) or (self.score_card is None):
            raise NotFittedError("RiskScore model has not been fitted yet")
        proba = self.predict_proba(X)

        return (proba[:, 1] >= proba_threshold).astype(int)

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        """Function that outputs probability of positive class according to risk-score model

        Args:
            X (pandas.DataFrame): dataset of features

        Returns:
            ndarray of shape (n_samples, 2): probability of negative and positive class in each column resp.
        """
        if (self.feature_point_card is None) or (self.score_card is None):
            raise NotFittedError("RiskScore model has not been fitted yet")

        # TODO have the same behaviour as predict proba from sklearn
        # TODO check if full numpy approach is faster
        df_score_card = self.score_card.copy().T.set_index("SCORE")
        list_features = self.feature_point_card["binary_feature"].values
        # TODO check that X only has selected featureso r just that it contains them?

        dataframe_schema = pa.DataFrameSchema(
            {
                c: pa.Column(checks=[pa.Check.isin([0, 1, 0.0, 1.0])])
                for c in list_features
            },
            strict=False,  # Disable check of other columns in the dataframe
        )

        dataframe_schema.validate(X)

        X_selected_features = X[list_features]

        points = self.feature_point_card[self._POINT_COL].values
        X_total_points = np.matmul(X_selected_features.values, points)
        proba = df_score_card.loc[X_total_points, "_RISK_FLOAT"].values.reshape(-1, 1)

        return np.concatenate([1 - proba, proba], axis=1)

    def summary(self) -> None:
        if (self.feature_point_card is None) or (self.score_card is None):
            raise NotFittedError("RiskScore model has not been fitted yet")
        feature_point_summary = (
            self.feature_point_card[[self._DESCRIPTION_COL, self._POINT_COL]]
            .sort_values(by=self._POINT_COL)
            .copy()
        )

        empty_column_name = " "
        feature_point_summary[empty_column_name] = "..."
        feature_point_summary.iloc[1:, -1] = "+ " + feature_point_summary.iloc[1:, -1]
        feature_point_summary.iloc[0, -1] = " " + feature_point_summary.iloc[0, -1]
        additional_display_rows = pd.DataFrame(
            data=[
                # [None, None, None],
                [" ", "SCORE=", " ..."]
            ],
            columns=feature_point_summary.columns,
            index=[" "],
        )

        feature_point_summary = pd.concat(
            [feature_point_summary, additional_display_rows], axis=0, ignore_index=False
        )

        feature_point_summary.index.name = self._FEATURE_COL

        print("======================")
        print("| FEATURE-POINT CARD |")
        print("======================")
        print(feature_point_summary.to_markdown())
        print()
        print("=======================================")
        print("=======================================")
        print()
        print("======================")
        print("|     SCORE CARD     |")
        print("======================")
        print(self.score_card.loc[["SCORE", "RISK"], :].to_markdown(headers="firstrow"))


class RiskScore(_BaseRiskScore):
    """
    Risk score model based on a binarizer, ranking of features, exhaustive enumeration with a maximization metric and a calibration method for probabilities.

    This class is a child class of _BaseRiskScore. It implements the fit method that creates the feature-point card and score card attribute.
    It computes them by binarizing features based on a given binarizer, ranking binary features and doing an exhaustive enumeration of features combination.
    It performs a hierarchical optimization:
    1) first it optimizes a metric based on scores. ROC AUC and/or PR AUC are good candidates as it is only based on the ranking of samples, compared to logloss which neeeds proper probabilities.
    The optimization is done by enumerating all possible selection of points for all combinations of binary feature and selecting the combination with the best value on the chosen metric;
    2) once the binary feature combination is chosen with the corresponding interger points, the logloss is optimized for each possible sum of points.

    Attributes
    ----------
    max_number_binaries_by_features : int
        maximum number of binary features to compute by feature
    nb_max_features: int
        maximum number of binary features to select
    min_point_value : int
        minimum points to assign for a binary feature
    max_point_value: ExplainableBoostingClassifier
        maximum points to assign for a binary feature
    binarizer:
        binarizer object that transforms continuous and categorical features into binary features



    Methods
    -------
    fit(self, X, y):
        function creating the feature-point card and score card attributes

    From _BaseRiskScore:

    @staticmethod
    _predict_proba_score(score, intercept):
        computes the value of the logistic function at score+intercept value

    predict(self, X, proba_threshold=0.5):
        function that predicts the positive/negative class by using a threshold on the probability given by the model

    predict_proba(self, X):
        function that predicts the probability of the positive class

    summary(self):
        function that prints the feature-point card and score card of the model
    """

    def __init__(
        self,
        binarizer: BinarizerProtocol,
        nb_max_features: int = 4,
        min_point_value: int = -2,
        max_point_value: int = 3,
        nb_additional_features: int = 4,
        ranker: Ranker = OMPRank(),
        calibrator: Calibrator = VanillaCalibrator(),
        enumeration_maximization_metric=fast_numba_auc,
    ):
        """
        Args:
            nb_max_features (int): Number of maximum binary features to be selected
            min_point_value (int, optional): Minimum point assigned to a binary feature. Defaults to -2.
            max_point_value (int, optional): Maximum point assigned to a binary feature. Defaults to 3.
            max_number_binaries_by_features (int, optional): Maximum number of binary features by original feature. Defaults to 3.
            nb_additional_features (int, optional): Number of additional features to consider for the binary feature selection. Defaults to 3.
            ranker (Ranker): Ranker object to rank binary features, see ranking.py
            calibrator (Calibrator): Calibrator object to define probabilities, see calibration.py
            enumeration_maximization_metric: maximization function used for enumeration.
                This function needs to compute a maximization metric based on 2 arguments in this order :
                1) numpy array containing the binary target
                2) numpy array containing the predicted probability or score
        """
        super().__init__(nb_max_features, min_point_value, max_point_value)

        # TODO Check value
        self.binarizer: BinarizerProtocol = binarizer
        self.nb_additional_features: int = int(nb_additional_features)
        self.ranker = ranker
        self.calibrator = calibrator
        self.enumeration_maximization_metric = enumeration_maximization_metric

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        X_calib: pd.DataFrame = None,
        y_calib: pd.Series = None,
        categorical_features="auto",
        fit_binarizer=True,
    ):
        """Function that search best parameters (choice of binary features, points and probabilities) of a risk score model.


        It computes them by binarizing features based on EBM, ranking binary features and doing an exhaustive enumeration of features combination.
        It performs a hierarchical optimization:
        1) first it optimizes ROC AUC and/or PR AUC by selecting points for all combinations of selected binary feature.
        The selection of binary features combination is done by:
            a) ranking the binary features,
            b) taking the top features according to the ranking,
            c) enumerate all combinations of binary features,
            d) enumerate all point assignment for each combination of binary features generated.
            e) choose the best combination of binary feature and point based on a ranking metric
        Different ranking techniques are available for step b) :
        LogOddsDensity, DiverseLogOddsDensity, CumulativeMetric, BordaRank, LassoPathRank, LarsPathRank, OMPRank, FasterRiskRank
        Everyone can implement its own ranking technique, given that the customized ranker class implements the Ranker class

        2) once the binary feature combination is chosen with the corresponding interger points, the logloss is optimized for each possible sum of points.
        The logloss optimization can be done on a different dataset (X_calib, y_calib).
        It can be done with a vanilla mode (preferred when train or calibration set is big enough), or a bootstrapped mode to avoid overfitting when there are few samples.
        These logloss optimizer with more details can be found in the calibration script of the package.
        Everyone can implement its own calibration technique, given that the customized calibrator class implements the Calibrator class

        Args:
            X (pandas.DataFrame): Dataset of features to fit the risk score model on
            y (pandas.Series): Target binary values
            X_calib (pandas.DataFrame): Dataset of features to calibrate probabilities on
            y_calib (pandas.Series): Target binary values for calibration
            categorical_features: list of categorical features for the binarizer
            fit_binarizer: boolean to indicate the binarizer should be fitted or not

        """
        start_time = time.time()

        # Binarize the features with the AutoBinarizer class
        if fit_binarizer:
            self.binarizer.fit(X, y, categorical_features=categorical_features)

        # TODO : check columns of df_info
        df_info = self.binarizer.get_info()

        # Prepare calibration set
        if X_calib is None:
            X_calib = X.copy()
            y_calib = y.copy()

        # Binarize the features with the previously fitted binarizer
        X_binarized = self.binarizer.transform(X)
        X_calib_binarized = self.binarizer.transform(X_calib)

        fit_transform_time = time.time() - start_time

        # Rank the binary feature by likeliness to be important for the risk score model

        df_ranker = df_info[["log_odds", "density", "feature"]].copy()

        df_rank = self.ranker.compute_ranking_features(
            df=df_ranker,
            X_binarized=X_binarized,
            y=y,
            nb_steps=(self.nb_additional_features + self.nb_max_features),
        )

        df_info = df_info.merge(
            df_rank, right_index=True, left_index=True, how="left", validate="1:1"
        )
        df_info = df_info.sort_values(by="rank", ascending=True)

        pool_top_features = df_info.index[
            : self.nb_max_features + self.nb_additional_features
        ]

        # Compute bounds for points for each feature to reduce optimization space
        # -> negative points for negative log odds
        # -> positive points for positive log odds
        df_sense = df_info.iloc[
            : self.nb_max_features + self.nb_additional_features
        ].copy()

        df_sense.loc[:, "lower_bound_point"] = np.where(
            df_sense["log_odds"] > 0,
            min([1, self.max_point_value]),
            self.min_point_value,
        )

        df_sense.loc[:, "upper_bound_point"] = np.where(
            df_sense["log_odds"] > 0,
            self.max_point_value,
            max([-1, self.min_point_value]),
        )

        # Define all possible integer values for each binary feature
        dict_point_ranges = {
            f: {
                "all_points": np.arange(
                    df_sense.loc[f, "lower_bound_point"],
                    df_sense.loc[f, "upper_bound_point"] + 1e-3,
                    1,
                ),
            }
            for f in pool_top_features
        }

        # For all combinations of nb_max_feature from the set of selected binary features
        nb_combi = len(
            list(itertools.combinations(pool_top_features, self.nb_max_features))
        )

        # Max number of point combination
        dim_max_point_combination = (
            max(len(dict_point_ranges[f]["all_points"]) for f in pool_top_features)
            ** self.nb_max_features
        )

        # Compute cube of all points combination :
        # number of combinations of self.nb_max_features feature among the selected pool of binary features x cardinality of feature pool
        cube = np.zeros(
            shape=(nb_combi * dim_max_point_combination, len(pool_top_features)),
            dtype=np.float16,
        )

        idx = 0
        # for each feature combination, compute point combinations
        for _, top_features in enumerate(
            itertools.combinations(pool_top_features, self.nb_max_features)
        ):
            # Gather all point ranges for each binary feature
            # if the feature is not in the current selected combination, the only point possibility is 0
            all_points_by_feature = [
                dict_point_ranges[f]["all_points"] if f in top_features else [0]
                for f in pool_top_features
            ]

            # Compute the cartesian product of all possible point values for each feature
            # This creates a nxd matrix with n being the number of combinations of points for each binary feature
            # and d being the total number of binary features in pool_top_features.
            # for each line, there are only self.nb_max_features non zero point values.
            all_points_possibilities = np.array(
                list(itertools.product(*all_points_by_feature))  # type: ignore
            )

            # Put this points combination into the global enumeration of points
            cube[
                idx : idx + all_points_possibilities.shape[0], :
            ] = all_points_possibilities
            idx += all_points_possibilities.shape[0]

        # Use dask to compute the metric to optimize on all points combination

        # Dask array for the global enumeration of combinations
        # TODO better design of chunk size
        dask_cube = da.from_array(cube, chunks=((50, len(pool_top_features))))

        # Dask array for the dataset of selected binary features
        dataset_transpose = X_binarized[list(pool_top_features)].values.T
        dask_dataset_T = da.from_array(
            dataset_transpose, chunks=dataset_transpose.shape
        )

        # 1Dfication of maximization metric function

        enumeration_optimization_metric_1d = lambda y_score: [
            self.enumeration_maximization_metric(y.values, y_score)
        ]

        # computation of score for all samples for all enumerated points combination
        matmul = da.matmul(dask_cube, dask_dataset_T)

        # computation of the maximum value on the defined enumeration metric
        idx_max = da.argmax(
            da.apply_along_axis(
                enumeration_optimization_metric_1d,
                axis=1,
                arr=matmul,
                shape=(1,),
                dtype=matmul.dtype,
            )
        ).compute()

        # Getting the points combination with the maximumu value
        best_points = cube[idx_max, :]
        best_feature_and_point_selection = [
            (point, f)
            for point, f in zip(best_points, list(pool_top_features))
            if point != 0
        ]

        best_scenario, best_feature_selection = zip(*best_feature_and_point_selection)
        best_feature_selection = list(best_feature_selection)

        # Probabilities computation

        df_calibration = pd.DataFrame(columns=["target", "score"])
        df_calibration["target"] = y_calib.values  # type: ignore
        df_calibration["score"] = np.matmul(
            X_calib_binarized[best_feature_selection].values, best_scenario
        )

        min_sum_point = np.sum(np.clip(best_scenario, a_max=0, a_min=None))
        max_sum_point = np.sum(np.clip(best_scenario, a_min=0, a_max=None))

        df_reordering = self.calibrator.calibrate(
            df_calibration, min_sum_point, max_sum_point
        )

        # Build feature-point card
        self.feature_point_card = pd.DataFrame(index=best_feature_selection)
        self.feature_point_card[self._POINT_COL] = best_scenario
        self.feature_point_card[self._POINT_COL] = self.feature_point_card[
            self._POINT_COL
        ].astype(int)
        self.feature_point_card["binary_feature"] = self.feature_point_card.index.values
        features_column = self.feature_point_card.merge(
            df_info["feature"], left_index=True, right_index=True
        )["feature"]
        self.feature_point_card[self._FEATURE_COL] = features_column

        self.feature_point_card = self.feature_point_card.set_index(self._FEATURE_COL)
        self.feature_point_card[self._DESCRIPTION_COL] = self.feature_point_card[
            "binary_feature"
        ].values

        possible_scores = list(df_reordering.index)
        possible_risks = df_reordering["sorted_proba"]
        # Compute nice display of probabilities
        possible_risks_pct = [f"{r:.2%}" for r in possible_risks]

        # Assign dataframe to score_card attribute
        self.score_card = pd.DataFrame(
            index=["SCORE", "RISK", "_RISK_FLOAT"],
            data=[possible_scores, possible_risks_pct, possible_risks],
        )

        self.score_card.loc["SCORE", :] = self.score_card.loc["SCORE", :].astype(int)
        self._total_fit_time = time.time() - start_time
        self._model_fit_time = self._total_fit_time - fit_transform_time

        return self

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        X_binarized = self.binarizer.transform(X)
        return super().predict_proba(X_binarized)


class EBMRiskScore(RiskScore):
    """
    Risk score model based on a EBMbinarizer, ranking of features, exhaustive enumeration with a maximization metric and a calibration method for probabilities.

    This class is a child class of _BaseRiskScore. It implements the fit method that creates the feature-point card and score card attribute.
    It computes them by binarizing features based on a given binarizer, ranking binary features and doing an exhaustive enumeration of features combination.
    It performs a hierarchical optimization:
    1) first it optimizes a metric based on scores. ROC AUC and/or PR AUC are good candidates as it is only based on the ranking of samples, compared to logloss which neeeds proper probabilities.
    The optimization is done by enumerating all possible selection of points for all combinations of binary feature and selecting the combination with the best value on the chosen metric;
    2) once the binary feature combination is chosen with the corresponding interger points, the logloss is optimized for each possible sum of points.

    Attributes
    ----------
    max_number_binaries_by_features : int
        maximum number of binary features to compute by feature
    nb_max_features: int
        maximum number of binary features to select
    min_point_value : int
        minimum points to assign for a binary feature
    max_point_value: ExplainableBoostingClassifier
        maximum points to assign for a binary feature
    binarizer:
        binarizer object that transforms continuous and categorical features into binary features



    Methods
    -------
    fit(self, X, y):
        function creating the feature-point card and score card attributes



    From _BaseRiskScore:

    @staticmethod
    _predict_proba_score(score, intercept):
        computes the value of the logistic function at score+intercept value

    predict(self, X, proba_threshold=0.5):
        function that predicts the positive/negative class by using a threshold on the probability given by the model

    predict_proba(self, X):
        function that predicts the probability of the positive class

    summary(self):
        function that prints the feature-point card and score card of the model
    """

    def __init__(
        self,
        nb_max_features: int = 4,
        min_point_value: int = -2,
        max_point_value: int = 3,
        max_number_binaries_by_features: int = 3,
        keep_negative: bool = True,
        nb_additional_features: Optional[int] = 4,
        ranker: Ranker = OMPRank(),
        calibrator: Calibrator = VanillaCalibrator(),
        enumeration_maximization_metric=fast_numba_auc,
    ):
        """
        Args:
            nb_max_features (int): Number of maximum binary features to be selected
            min_point_value (int, optional): Minimum point assigned to a binary feature. Defaults to -2.
            max_point_value (int, optional): Maximum point assigned to a binary feature. Defaults to 3.
            max_number_binaries_by_features (int, optional): Maximum number of binary features by original feature. Defaults to 3.
            nb_additional_features (int, optional): Number of additional features to consider for the binary feature selection. Defaults to 3.
            ranker (Ranker): Ranker object to rank binary features, see ranking.py
            calibrator (Calibrator): Calibrator object to define probabilities, see calibration.py
            enumeration_maximization_metric: maximization function used for enumeration.
                This function needs to compute a maximization metric based on 2 arguments in this order :
                1) numpy array containing the binary target
                2) numpy array containing the predicted probability or score

        """
        super().__init__(
            binarizer=EBMBinarizer(
                max_number_binaries_by_features=max_number_binaries_by_features,
                keep_negative=keep_negative,
            ),
            nb_max_features=nb_max_features,
            min_point_value=min_point_value,
            max_point_value=max_point_value,
            nb_additional_features=nb_additional_features,
            ranker=ranker,
            calibrator=calibrator,
            enumeration_maximization_metric=enumeration_maximization_metric,
        )



import warnings

import pytest
from pandera.errors import SchemaError
from sklearn.exceptions import NotFittedError

from scorepyo.binarizers import EBMBinarizer
from scorepyo.exceptions import (
    MissingColumnError,
    NegativeValueError,
    NonBooleanValueError,
    NonIntegerValueError,
)

# from pytest_lazyfixture import lazy_fixture


def test_binarizer_integer_param():
    with pytest.raises(NonIntegerValueError):
        EBMBinarizer(max_number_binaries_by_features=3.5)


def test_binarizer_positive_param():
    with pytest.raises(NegativeValueError):
        EBMBinarizer(max_number_binaries_by_features=-1)


@pytest.mark.parametrize(
    "keep_negative_value",
    [2, "True", "foo", [1, 2], [True], 1, 0],
)
def test_binarizer_boolean_param(keep_negative_value):
    with pytest.raises(NonBooleanValueError):
        EBMBinarizer(keep_negative=keep_negative_value)


def test_missing_categorical_columns(mixed_features, binary_target):
    binarizer = EBMBinarizer()
    with pytest.raises(MissingColumnError):
        binarizer.fit(mixed_features, binary_target, categorical_features=["D"])


def test_missing_to_exclude_columns(mixed_features, binary_target):
    binarizer = EBMBinarizer()
    with pytest.raises(MissingColumnError):
        binarizer.fit(mixed_features, binary_target, to_exclude_features=["D"])


def test_detect_categorical_columns(mixed_features, binary_target):
    binarizer = EBMBinarizer()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
        )
        binarizer.fit(mixed_features, binary_target, categorical_features="auto")

    assert binarizer._categorical_features == ["C"]


def test_not_fitted(mixed_features):
    binarizer = EBMBinarizer()
    with pytest.raises(NotFittedError):
        binarizer.transform(mixed_features)


def test_different_dataframe(mixed_features, continuous_features, binary_target):
    binarizer = EBMBinarizer()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
        )
        binarizer.fit(mixed_features, binary_target, categorical_features="auto")
    with pytest.raises(SchemaError):
        binarizer.transform(continuous_features)


@pytest.mark.parametrize(
    "max_plateau",
    [2, 3, 5, 10],
)
def test_number_plateaus(breast_cancer_features, breast_cancer_target, max_plateau):
    X, y = breast_cancer_features, breast_cancer_target
    binarizer = EBMBinarizer(
        max_number_binaries_by_features=max_plateau, keep_negative=True
    )
    binarizer.fit(X, y)
    X_binarized = binarizer.transform(X)
    assert len(X_binarized.columns) == max_plateau * len(X.columns)


def test_define_categorical_columns(mixed_features, binary_target):
    binarizer = EBMBinarizer()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
        )
        binarizer.fit(mixed_features, binary_target, categorical_features=["C"])
    assert binarizer._categorical_features == ["C"]
    X_binarized = binarizer.transform(mixed_features)
    # test if all categories of C are in the new columns
    assert all([(c in X_binarized.columns) for c in ["C_1", "C_2"]])


@pytest.mark.parametrize(
    "excluded_columns",
    [["A"], ["B"], ["C"], ["A", "B"], ["A", "C"], ["B", "C"], ["A", "B", "C"]],
)
def test_define_exclude_columns(mixed_features, binary_target, excluded_columns):
    binarizer = EBMBinarizer()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
        )
        binarizer.fit(
            mixed_features, binary_target, to_exclude_features=excluded_columns
        )
    print(binarizer._to_exclude_features)
    print(excluded_columns)

    assert binarizer._to_exclude_features == excluded_columns

    X_binarized = binarizer.transform(mixed_features)
    print(X_binarized.columns)
    # test that all columns excluded from binarizer are left as is
    assert all([(c in X_binarized.columns) for c in excluded_columns])

    # test that the new dataframe has the same values as in the original
    assert mixed_features[excluded_columns].equals(X_binarized[excluded_columns])


def test_keep_negative(breast_cancer_features, breast_cancer_target):
    binarizer = EBMBinarizer(keep_negative=False)
    binarizer.fit(breast_cancer_features, breast_cancer_target)
    binarizer.transform(breast_cancer_features)
    df_info = binarizer.get_info()
    print(df_info)
    mask_no_intercept = df_info.index != "intercept"
    mask_nan = ~df_info["log_odds"].isna()
    assert df_info[mask_no_intercept & mask_nan]["log_odds"].astype(float).min() >= 0



# import warnings

import pytest

# from pandera.errors import SchemaError
from sklearn.exceptions import NotFittedError

from scorepyo.exceptions import (
    MinPointOverMaxPointError,
    NegativeValueError,
    NonIntegerValueError,
)
from scorepyo.models import EBMRiskScore, _BaseRiskScore


def test_binarizer_nb_max_features():
    with pytest.raises(NegativeValueError):
        EBMRiskScore(nb_max_features=-1)

    with pytest.raises(NonIntegerValueError):
        EBMRiskScore(nb_max_features=1.2)


def test_min_point_value():
    with pytest.raises(NonIntegerValueError):
        EBMRiskScore(min_point_value=1.2)

    with pytest.raises(NonIntegerValueError):
        EBMRiskScore(min_point_value=-2.7)


def test_max_point_value():
    with pytest.raises(NonIntegerValueError):
        EBMRiskScore(max_point_value=1.2)

    with pytest.raises(NonIntegerValueError):
        EBMRiskScore(max_point_value=-2.7)


def test_min_over_max_point_value():
    with pytest.raises(MinPointOverMaxPointError):
        EBMRiskScore(min_point_value=3, max_point_value=1)


# @pytest.mark.parametrize(
#     "erroneous_column_names",
#     [
#         ["confeature", "binary_feature"],
#         ["feature", "binary_confeature"],
#         ["confeature", "binary_confeature"],
#     ],
# )
# def test_schema_df_info(df_info_binary_features, erroneous_column_names):
#     erroneous_df_info = df_info_binary_features.copy()
#     erroneous_df_info.columns = erroneous_column_names
#     with pytest.raises(SchemaError):
#         EBMRiskScore(df_info=erroneous_df_info)


def test_base_score_card_fit(binary_features, binary_target):
    with pytest.raises(NotImplementedError):
        score_card = _BaseRiskScore()
        score_card.fit(binary_features, binary_target)


def test_optuna_score_card_predict(binary_features):
    with pytest.raises(NotFittedError):
        score_card = EBMRiskScore()
        score_card.predict(binary_features)


def test_optuna_score_card_predict_proba(binary_features):
    with pytest.raises(NotFittedError):
        score_card = EBMRiskScore()
        score_card.predict_proba(binary_features)


def test_optuna_score_card_summary():
    with pytest.raises(NotFittedError):
        score_card = EBMRiskScore()
        score_card.summary()



"""Classes for binary features rankers
"""

from abc import ABC, abstractmethod
from typing import Any, List

import numpy as np
import pandas as pd
import pandera as pa
from fasterrisk.fasterrisk import RiskScoreOptimizer
from sklearn.linear_model import OrthogonalMatchingPursuit, lars_path, lasso_path


class Ranker(ABC):
    """Base class for binary features ranker

    Child classes must implement the _compute_ranking_features method.

    """

    def compute_ranking_features(
        self, df: pd.DataFrame, *args: Any, **kwargs: Any
    ) -> pd.Series:
        series_ranking = self._compute_ranking_features(df, *args, **kwargs)

        series_ranking_schema = pa.SeriesSchema(
            # dtype=pa.Int(),
            name="rank",
            checks=[
                pa.Check.isin(range(len(df))),
                pa.Check(
                    lambda x: int(x) == x, element_wise=True
                ),  # check that type is an int (any bit number int32, int64, etc.)
            ],
        )
        series_ranking_schema.validate(series_ranking)
        return series_ranking

    @abstractmethod
    def _compute_ranking_features(
        # self, df: pd.DataFrame,
        *args: Any,
        **kwargs: Any,
    ) -> pd.Series:
        raise NotImplementedError


class LogOddsDensity(Ranker):
    """Binary features ranker based on logodds contribution of binary feature,
    and reduced importance of density of the binary feature (number of samples with a positive value on the binary feature)



    Child class of Ranker.
    """

    def __init__(self, **kwargs):  # pylint: disable=W0613
        ...

    def _compute_ranking_features(
        self, df: pd.DataFrame, *args: Any, **kwargs: Any
    ) -> pd.Series:  # pylint disable=W0613

        """
        This method ranks the binary features based on the product of :
         - logodds contribution of the binary feature
         - reduced importance of density

        This prevents having a too large emphasis on a binary feature with a large density and low log odd contribution
        Args:
            df (pd.DataFrame): DataFrame with logodds and density information, with the binary feature name as an index

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """

        dataframe_schema = pa.DataFrameSchema(
            {"log_odds": pa.Column(nullable=True), "density": pa.Column(nullable=True)},
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()

        # The density importance is reduced by taking a power of 0.95
        # TODO Put the power in the argument of the class
        df_["abs_contribution"] = df_["log_odds"].abs() * df_["density"].fillna(
            0
        ).astype(int).pow(0.95)
        df_["rank"] = df_["abs_contribution"].rank(
            method="min", na_option="bottom", ascending=False
        )
        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class DiverseLogOddsDensity(Ranker):
    """Binary features ranker based on logodds contribution of binary feature,
    and density of the binary feature (number of samples with a positive value on the binary feature)

    In order to diversify the ranking, a small number of binary features coming from the same origin will be at the top of the ranking


    Child class of Ranker.
    """

    def __init__(self, rank_diversity: int = 1, **kwargs):  # pylint: disable=W0613
        """
        Args:
            rank_diversity (int, optional): Indicates the number of additional binary features from the same origin to include in the ranking. Defaults to 1.
        """
        # TODO check parameters
        self.rank_diversity = rank_diversity

    def _compute_ranking_features(
        self, df: pd.DataFrame, *args: Any, **kwargs: Any
    ) -> pd.Series:
        """This method ranks the binary features based on the product of :
         - logodds contribution of the binary feature
         - importance of density
         - origin of binary features

         If several binary features are at the top of the ranking, only self.rank_diversity+1 will be kept

        Args:
            df (pd.DataFrame): DataFrame with logodds, density and origin information, with the binary feature name as an index

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
                "density": pa.Column(nullable=True),
                "feature": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()

        df_["log_odds_sum"] = df_["log_odds"].abs() * df_["density"].fillna(0)

        df_["absolute_log_odds"] = df_["log_odds"].abs()

        # keep first kth binary feature of a feature and mark it as priority
        idx_up_to_kth = (
            df_.sort_values(
                ["log_odds_sum", "absolute_log_odds"], ascending=[False, False]
            )
            .reset_index()
            .groupby("feature")["binary_feature"]
            .nth(n=list(range(self.rank_diversity + 1)))
            .values
        )

        idx_up_to_kth = [f for f in idx_up_to_kth if f != "intercept"]

        df_["prio"] = np.where(df_.index.isin(idx_up_to_kth), 1, 0)

        df_ = df_.sort_values(by=["prio", "log_odds_sum"], ascending=[False, False])
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class CumulativeMetric(Ranker):
    """Binary features ranker based on computing a metric on a growing number of binary features.

    This ranker initially sorts binary features depending on a specified ranker.
    Then it computes a classification metric by adding one log odd contribution of a binary feature at a time.
    The binary features are then ranked according to the incremental difference they made on the metric.

    Child class of Ranker
    """

    def __init__(self, metric, ranker: Ranker, **kwargs):  # pylint: disable=W0613
        """
        Args:
            metric (function): function taking 2 arguments: binary target and probabilities/scores
                this function should be maximized
            ranker (Ranker): original ranker
        """
        # TODO put a check on the metric function
        self.metric = metric
        self.ranker = ranker

    def _compute_ranking_features(
        self,
        df: pd.DataFrame,
        X_binarized: pd.DataFrame,
        y: pd.Series,
        *args: Any,
        **kwargs: Any,
    ) -> pd.Series:  # type: ignore
        """This ranker initially sorts binary features depending on a specified ranker.
        Then it computes a classification metric by adding one log odd contribution of a binary feature at a time.
        The binary features are then ranked according to the magnitude of the incremental difference they made on the metric.

        Args:
            df (pd.DataFrame): information dataframe, it should contain the log_odds contribution of binary feature
            X_binarized (pd.DataFrame): binary dataset
            y (pd.Series): binary target

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        # TODO add check on intercept in index
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()

        df_rank = self.ranker.compute_ranking_features(df_)

        df_ = df_.merge(
            df_rank, right_index=True, left_index=True, how="left", validate="1:1"
        )

        # Sort binary features according to a ranker
        df_ = df_.sort_values(by="rank", ascending=True)

        features_sorted = pd.Index([f for f in df_.index if f != "intercept"])

        # Compute log_odds vector of each binary feature
        log_odds_vector = df_.loc[features_sorted, "log_odds"].values

        log_odd_intercept = df_.loc["intercept", "log_odds"]

        log_odds_vector = log_odds_vector.reshape(-1, 1)

        # Add logodd intercept to all log_odds
        log_odds_vector = log_odds_vector + log_odd_intercept

        # Duplicate log_odds vector as many times as there are binary features
        log_odds_matrix = np.tile(log_odds_vector, reps=(1, len(features_sorted)))

        # Decreasingly put to 0 log odd the 2nd to nth binary feature, then 3rd to nth, etc.
        for i in range(len(features_sorted)):
            log_odds_matrix[i + 1 :, i] = 0

        # Compute cumulative logodds by introducing one binary feature on top of each other
        cumulative_log_odds = np.matmul(
            X_binarized[features_sorted].values, log_odds_matrix
        )

        def metric_function_1d(y_proba):
            return self.metric(y.astype(float).values, np.array(y_proba, dtype=float))

        # Compute metric on cumulative logodds
        cumulative_metric = np.apply_along_axis(
            metric_function_1d, axis=0, arr=cumulative_log_odds
        )

        # Compute the incremental difference between a metric and the previous value
        diff_cumulative_metric = [
            t - s for s, t in zip(cumulative_metric, cumulative_metric[1:])
        ]
        diff_cumulative_metric = [diff_cumulative_metric[0]] + diff_cumulative_metric

        # Sort the binary features based on the magnitude of metric difference in a decreasing fashion
        idx_sorted_features_metric_diff = np.argsort(np.array(diff_cumulative_metric))[
            ::-1
        ]

        ranked_idx = features_sorted[idx_sorted_features_metric_diff]
        ranked_idx = list(ranked_idx) + ["intercept"]

        df_ = df_.loc[ranked_idx, :]
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class BordaRank(Ranker):
    """Based on a list of Ranker, computes the Borda rank of each binary feature based on all rankers.

    Borda rank : https://en.wikipedia.org/wiki/Borda_count

    Child class of ranker
    """

    def __init__(self, list_ranker: List[Ranker], **kwargs):  # pylint: disable=W0613
        self.list_ranker = list_ranker.copy()

    def _compute_ranking_features(
        self, df: pd.DataFrame, *args: Any, **kwargs: Any
    ) -> pd.Series:
        """Based on a list of Ranker, computes the Borda rank of each binary feature based on all rankers.

        Borda rank : https://en.wikipedia.org/wiki/Borda_count

        Args:
            df (pd.DataFrame): DataFrame of logodds, density, features and binary features information needed for all rankers

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        df_ = df.copy()

        df_list_ranker = pd.DataFrame(
            columns=[str(i) for i in range(len(self.list_ranker))], index=df_.index
        )
        for i, ranker in enumerate(self.list_ranker):
            df_list_ranker[str(i)] = ranker.compute_ranking_features(df_, **kwargs)

        df_ = df_.merge(
            df_list_ranker.sum(axis=1).to_frame(name="rank"),
            left_index=True,
            right_index=True,
            validate="1:1",
            how="left",
        )

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class LassoPathRank(Ranker):
    """Binary feature ranker based on lasso path.

    Lasso path stores the coefficient values of features along different values of the regularization parameters,
    of a Lasso regression.
    Based on this, the ranker selects the lasso path step where the number of binary features with non-zero coefficient
    is equal to the specified target.

    For more info on Lasso path : https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.lasso_path.html



    Child class of Ranker
    """

    def __init__(self, **kwargs):  # pylint: disable=W0613
        ...

    def _compute_ranking_features(
        self,
        df: pd.DataFrame,
        X_binarized: pd.DataFrame,
        y: pd.Series,
        nb_steps: int,
        **kwargs,
    ) -> pd.Series:
        """This functions returns the rank according to the lasso path
        Args:
            df (pd.DataFrame): Information dataframe on logodds, density and binary feature name
            X_binarized (pd.DataFrame): Binary features
            y (pd.Series): Binary target
            nb_steps (int): Number of non zero coefficient to identify the step in the lasso path

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
                "density": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()
        df_["log_odds_sum"] = df_["log_odds"].abs() * df_["density"].fillna(0)

        _, coefs_lasso, _ = lasso_path(X_binarized.values, y, verbose=False)
        coefs_lasso_updated = [coefs_lasso[:, 0]]
        for i in range(1, coefs_lasso.shape[1]):
            previous_column_zero_count = np.sum(np.where(coefs_lasso_updated[-1] == 0))
            current_column_zero_count = np.sum(np.where(coefs_lasso[:, i] == 0))
            if current_column_zero_count < previous_column_zero_count:
                coefs_lasso_updated.append(coefs_lasso[:, i])
        coefs_lasso_updated_ = np.array(coefs_lasso_updated).T
        lasso_selected_features = [
            c
            for v, c in zip(coefs_lasso_updated_[:, nb_steps], X_binarized.columns)
            if v != 0
        ]

        df_["lasso_selected_feature"] = np.where(
            df_.index.isin(lasso_selected_features), 1, 0
        )

        df_ = df_.sort_values(
            by=["lasso_selected_feature", "log_odds_sum"], ascending=[False, False]
        )
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class LarsPathRank(Ranker):
    """Binary feature ranker based on LARS lasso path.

    LARS lasso path is the coefficient values of features along different values of the regularization parameters,
    of a LARS Lasso regression.
    Based on this, the ranker selects the LARS lasso path step where the number of binary features with non-zero coefficient
    is equal to the specified target.

    For more info on LARS Lasso path : https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.lars_path.html#sklearn.linear_model.lars_path


    Child class of Ranker
    """

    def __init__(self, **kwargs):  # pylint: disable=W0613
        ...

    def _compute_ranking_features(
        self,
        df: pd.DataFrame,
        X_binarized: pd.DataFrame,
        y: pd.Series,
        nb_steps: int,
        **kwargs,
    ) -> pd.Series:
        """This function returns binary features rank based on LARS Lasso path

        Args:
            df (pd.DataFrame): Information dataframe on logodds, density and binary feature name
            X_binarized (pd.DataFrame): Binary features
            y (pd.Series): Binary target
            nb_steps (int): Number of non zero coefficient to identify the step in the LARS lasso path

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
                "density": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()
        df_["log_odds_sum"] = df_["log_odds"].abs() * df_["density"].fillna(0)

        _, _, coefs_lars = lars_path(
            X_binarized.astype(float).values,
            y.astype(float).values,
            method="lasso",
            verbose=False,
        )
        lars_selected_features = [
            c for v, c in zip(coefs_lars[:, nb_steps], X_binarized.columns) if v != 0
        ]

        df_["lars_selected_feature"] = np.where(
            df_.index.isin(lars_selected_features), 1, 0
        )

        df_ = df_.sort_values(
            by=["lars_selected_feature", "log_odds_sum"], ascending=[False, False]
        )
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class OMPRank(Ranker):
    """Binary feature ranker based on Orthogonal Matching Pursuit.

       From https://scikit-learn.org/stable/modules/linear_model.html#omp:
       " Orthogonal Matching Pursuit algorithm  approximates the fit of a linear model with constraints imposed on the number of non-zero coefficients (ie. the
    pseudo-norm).

       Based on this, the ranker selects the binary features selected by the OMP algorithm where the number of non zero coefficients is equal to the specified target.

       Child class of Ranker.
    """

    def __init__(self, **kwargs):  # pylint: disable=W0613
        ...

    def _compute_ranking_features(
        self,
        df: pd.DataFrame,
        X_binarized: pd.DataFrame,
        y: pd.Series,
        nb_steps: int,
        **kwargs,
    ) -> pd.Series:
        """This function returns binary features rank based on OMP

        Args:
            df (pd.DataFrame): Information dataframe on logodds, density and binary feature name
            X_binarized (pd.DataFrame): Binary features
            y (pd.Series): Binary target
            nb_steps (int): Number of non zero coefficient to identify the step in the LARS lasso path

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
                "density": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()
        df_["log_odds_sum"] = df_["log_odds"].abs() * df_["density"].fillna(0)

        omp = OrthogonalMatchingPursuit(n_nonzero_coefs=nb_steps, normalize=False)
        omp.fit(X_binarized, y)

        omp_selected_features = [
            c for v, c in zip(omp.coef_, X_binarized.columns) if v != 0
        ]

        df_["omp_selected_features"] = np.where(
            df_.index.isin(omp_selected_features), 1, 0
        )

        df_ = df_.sort_values(
            by=["omp_selected_features", "log_odds_sum"], ascending=[False, False]
        )
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)


class FasterRiskRank(Ranker):
    """Binary features ranker based on FasterRisk, another risk score model library.

    FasterRisk has its own algorithm to select candidate binary features.
    Based on their algorithm, we take the top features based on number of appearances in candidate models.
    Then we sort by the product of density and lod odds of binary features

    For more information of FasterRisk package and usage: https://fasterrisk.readthedocs.io/en/latest/

    Child class of Ranker
    """

    def __init__(
        self,
        parent_size=10,
        child_size=None,
        max_attempts=50,
        num_ray=20,
        lineSearch_early_stop_tolerance=1e-3,
        min_point_value=-2,
        max_point_value=3,
        nb_max_features=4,
        **kwargs,
    ):  # pylint: disable=W0613
        self.parent_size = parent_size
        self.child_size = child_size
        self.max_attempts = max_attempts
        self.num_ray = num_ray
        self.lineSearch_early_stop_tolerance = lineSearch_early_stop_tolerance
        self.min_point_value = min_point_value
        self.max_point_value = max_point_value
        self.nb_max_features = nb_max_features

    def _compute_ranking_features(
        self,
        df: pd.DataFrame,
        X_binarized: pd.DataFrame,
        y: pd.Series,
        nb_steps: int,
        **kwargs,
    ) -> pd.Series:
        # pylint disable=W0613
        """This function returns binary features rank based on number of appearances in FasterRisk candidate risk score models

        Args:
            df (pd.DataFrame): Information dataframe on logodds, density and binary feature name
            X_binarized (pd.DataFrame): Binary features
            y (pd.Series): Binary target
            nb_steps (int): Number of non zero coefficient to identify the step in the LARS lasso path

        Returns:
            pd.Series: Rank for each binary feature in a decreasing order of importance
        """
        dataframe_schema = pa.DataFrameSchema(
            {
                "log_odds": pa.Column(nullable=True),
                "density": pa.Column(nullable=True),
            },
            index=pa.Index(str, name="binary_feature"),
            strict=False,  # disable check of other columns in the dataframe
        )

        dataframe_schema.validate(df)

        df_ = df.copy()
        df_["log_odds_sum"] = df_["log_odds"].abs() * df_["density"].fillna(0)

        RiskScoreOptimizer_m = RiskScoreOptimizer(
            X=X_binarized.values,
            y=np.where(y == 0, -1, 1),
            lb=self.min_point_value,
            ub=self.max_point_value,
            k=self.nb_max_features,
            parent_size=self.parent_size,
            child_size=self.child_size,
            maxAttempts=self.max_attempts,
            num_ray_search=self.num_ray,
            lineSearch_early_stop_tolerance=self.lineSearch_early_stop_tolerance,
        )

        RiskScoreOptimizer_m.sparseLogRegModel_object.get_sparse_sol_via_OMP(
            k=RiskScoreOptimizer_m.k,
            parent_size=RiskScoreOptimizer_m.parent_size,
            child_size=RiskScoreOptimizer_m.child_size,
        )

        (
            beta0,
            betas,
            ExpyXB,
        ) = RiskScoreOptimizer_m.sparseLogRegModel_object.get_beta0_betas_ExpyXB()
        RiskScoreOptimizer_m.sparseDiversePoolLogRegModel_object.warm_start_from_beta0_betas_ExpyXB(
            beta0=beta0, betas=betas, ExpyXB=ExpyXB
        )

        (
            _,
            sparseDiversePool_betas,
        ) = RiskScoreOptimizer_m.sparseDiversePoolLogRegModel_object.get_sparseDiversePool(
            gap_tolerance=RiskScoreOptimizer_m.sparseDiverseSet_gap_tolerance,
            select_top_m=RiskScoreOptimizer_m.sparseDiverseSet_select_top_m,
            maxAttempts=RiskScoreOptimizer_m.sparseDiverseSet_maxAttempts,
        )
        features_sorted_by_occurence = sorted(
            [
                (f, v)
                for f, v in zip(
                    X_binarized.columns,
                    np.sum((sparseDiversePool_betas != 0), axis=0),
                )
            ],
            key=lambda t: t[1],
        )
        fasterrisk_chosen_features = [
            f for f, _ in features_sorted_by_occurence[-(nb_steps):]
        ]

        df_["fasterrisk_chosen_features"] = np.where(
            df_.index.isin(fasterrisk_chosen_features), 1, 0
        )

        df_ = df_.sort_values(
            by=["fasterrisk_chosen_features", "log_odds_sum"], ascending=[False, False]
        )
        df_["rank"] = list(range(len(df_)))
        df_["rank"] = df_["rank"] + 1

        index_without_intercept = [c for c in df_.index if c != "intercept"]
        return df_.loc[index_without_intercept, "rank"].astype(int)
